#pragma once

#define PT2_PI 3.14159265358979323846264338327950288
#define PT2_TWO_PI 6.28318530717958647692528676655900576

double pt2_sqrt(double x);
double pt2_tan(double x);
double pt2_cos(double x);
double pt2_sin(double x);
